/*
 * ringofframes.c
 * Ring Buffer utility..
 *
 * Original Author: Phil Burk, http://www.softsynth.com
 * Adapted by Roger B. Dannenberg
 *
 * This program uses the PortAudio Portable Audio Library.
 * For more information see: http://www.audiomulch.com/portaudio/
 * Copyright (c) 1999-2000 Ross Bencina and Phil Burk
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * Any person wishing to distribute modifications to the Software is
 * requested to send the modifications to the original developer so that
 * they can be incorporated into the canonical version.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ringofframes.h"
#include <string.h>

/***************************************************************************
 * Initialize FIFO.
 */
void RingOfFrames_Init( RingOfFrames *rbuf, long numFrames, long numBytesPerFrame, void *dataPtr )
{
    rbuf->bufferSize = numFrames;
    rbuf->frameSize = numBytesPerFrame;
    rbuf->buffer = (char *)dataPtr;
    RingOfFrames_Flush( rbuf );
}
/***************************************************************************
** Return number of frames available for reading. */
long RingOfFrames_GetReadAvailable( RingOfFrames *rbuf )
{
    // writeIndex is volatile, so get a copy
    long writeIndex = rbuf->writeIndex;
    long readIndex = rbuf->readIndex; // might as well copy this too
    if (writeIndex < readIndex) writeIndex += rbuf->bufferSize;
    return writeIndex - readIndex;
}
/***************************************************************************
** Return number of frames available for writing. */
long RingOfFrames_GetWriteAvailable( RingOfFrames *rbuf )
{
    // readIndex is volatile, so get a copy
    long readIndex = rbuf->readIndex;
    long writeIndex = rbuf->writeIndex; // might as well copy this too
    if (writeIndex >= readIndex) readIndex += rbuf->bufferSize;
    return readIndex - writeIndex - 1;
}

/***************************************************************************
** Clear buffer. Should only be called when buffer is NOT being read. */
void RingOfFrames_Flush( RingOfFrames *rbuf )
{
    rbuf->writeIndex = rbuf->readIndex = 0;
}

/***************************************************************************
** Get address of region(s) to which we can write data.
** If the region is contiguous, size2 will be zero.
** If non-contiguous, size2 will be the size of second region.
** Returns room available to be written or numFrames, whichever is smaller.
*/
long RingOfFrames_GetWriteRegions( RingOfFrames *rbuf, long numFrames,
                                 void **dataPtr1, long *sizePtr1,
                                 void **dataPtr2, long *sizePtr2 )
{
    long index = rbuf->writeIndex;
    long   available = RingOfFrames_GetWriteAvailable( rbuf );
    if( numFrames > available ) numFrames = available;
    *dataPtr1 = rbuf->buffer + rbuf->frameSize * index;
    /* Check to see if write is not contiguous. */
    if( (index + numFrames) > rbuf->bufferSize )
    {
        /* Write data in two frames that wrap the buffer. */
        long   firstHalf = rbuf->bufferSize - index;
        *sizePtr1 = firstHalf;
        *dataPtr2 = rbuf->buffer;
        *sizePtr2 = numFrames - firstHalf;
    }
    else
    {
        *sizePtr1 = numFrames;
        *dataPtr2 = NULL;
        *sizePtr2 = 0;
    }
    return numFrames;
}


/***************************************************************************
*/
long RingOfFrames_AdvanceWriteIndex( RingOfFrames *rbuf, long numFrames )
{
    /* assume numFrames < rbuf->bufferSize */
    long index = rbuf->writeIndex + numFrames;
    if (index >= rbuf->bufferSize) index -= rbuf->bufferSize;
    return rbuf->writeIndex = index;
}

/***************************************************************************
** Get address of region(s) from which we can read data.
** If the region is contiguous, size2 will be zero.
** If non-contiguous, size2 will be the size of second region.
** Returns room available to be written or numFrames, whichever is smaller.
*/
long RingOfFrames_GetReadRegions( RingOfFrames *rbuf, long numFrames,
                                void **dataPtr1, long *sizePtr1,
                                void **dataPtr2, long *sizePtr2 )
{
    long   index = rbuf->readIndex;
    long   available = RingOfFrames_GetReadAvailable( rbuf );
    if( numFrames > available ) numFrames = available;
    *dataPtr1 = rbuf->buffer + rbuf->frameSize * index;
    /* Check to see if read is not contiguous. */
    if( (index + numFrames) > rbuf->bufferSize )
    {
        /* Write data in two frames that wrap the buffer. */
        long firstHalf = rbuf->bufferSize - index;
        *sizePtr1 = firstHalf;
        *dataPtr2 = rbuf->buffer;
        *sizePtr2 = numFrames - firstHalf;
    }
    else
    {
        *sizePtr1 = numFrames;
        *dataPtr2 = NULL;
        *sizePtr2 = 0;
    }
    return numFrames;
}
/***************************************************************************
*/
long RingOfFrames_AdvanceReadIndex( RingOfFrames *rbuf, long numFrames )
{
    /* assume numFrames < rbuf->bufferSize */
    long index = rbuf->readIndex + numFrames;
    if (index >= rbuf->bufferSize) index -= rbuf->bufferSize;
    return rbuf->readIndex = index;
}

/***************************************************************************
** Return frames written. */
long RingOfFrames_Write( RingOfFrames *rbuf, void *data, long numFrames )
{
    long size1, size2, numWritten;
    void *data1, *data2;
    numWritten = RingOfFrames_GetWriteRegions( rbuf, numFrames, &data1, &size1, &data2, &size2 );
    memcpy( data1, data, rbuf->frameSize * size1 );
    if( size2 > 0 )
    {
        data = ((char *)data) + rbuf->frameSize * size1;
        memcpy( data2, data, rbuf->frameSize * size2 );
    }
    RingOfFrames_AdvanceWriteIndex( rbuf, numWritten );
    return numWritten;
}

/***************************************************************************
** Return frames read. */
long RingOfFrames_Read( RingOfFrames *rbuf, void *data, long numFrames )
{
    long size1, size2, numRead;
    void *data1, *data2;
    numRead = RingOfFrames_GetReadRegions( rbuf, numFrames, &data1, &size1, &data2, &size2 );
    memcpy( data, data1, rbuf->frameSize * size1 );
    if( size2 > 0 )
    {
        data = ((char *)data) + rbuf->frameSize * size1;
        memcpy( data, data2, rbuf->frameSize * size2 );
    }
    RingOfFrames_AdvanceReadIndex( rbuf, numRead );
    return numRead;
}
